from __future__ import annotations

from collections.abc import Mapping
from copy import copy
from typing import Protocol


class FieldMapping(Protocol):
    """Maps logical field names to AAMVA element ids (``field_for`` contract)."""

    def field_for(self, key: str) -> str: ...


class FieldMapper:
    """Default AAMVA field designators (aligned with Version 8-style layout)."""

    def __init__(self) -> None:
        self._fields: dict[str, str] = {
            "first_name": "DAC",
            "last_name": "DCS",
            "middle_name": "DAD",
            "expiration_date": "DBA",
            "issue_date": "DBD",
            "date_of_birth": "DBB",
            "gender": "DBC",
            "eye_color": "DAY",
            "height": "DAU",
            "street_address": "DAG",
            "city": "DAI",
            "state": "DAJ",
            "postal_code": "DAK",
            "drivers_license_id": "DAQ",
            "document_id": "DCF",
            "country": "DCG",
            "middle_name_truncation": "DDG",
            "first_name_truncation": "DDF",
            "last_name_truncation": "DDE",
            "street_address_supplement": "DAH",
            "hair_color": "DAZ",
            "place_of_birth": "DCI",
            "audit_information": "DCJ",
            "inventory_control_number": "DCK",
            "last_name_alias": "DBN",
            "first_name_alias": "DBG",
            "suffix_alias": "DBS",
            "suffix": "DCU",
            "weight": "DAW",
        }

    @property
    def fields(self) -> Mapping[str, str]:
        return self._fields

    @fields.setter
    def fields(self, new_fields: dict[str, str]) -> None:
        self._fields = new_fields

    def field_for(self, key: str) -> str:
        return self._fields.get(key, "")

    def copy_fields(self) -> dict[str, str]:
        return copy(self._fields)



from __future__ import annotations

from datetime import datetime

from aamva_parser.field_parser import FieldParser
from aamva_parser.license import License
from aamva_parser.parsed_license import ParsedLicense
from aamva_parser.regex_util import Regex
from aamva_parser.version_field_parsers import (
    VersionEightFieldParser,
    VersionElevenFieldParser,
    VersionFiveFieldParser,
    VersionFourFieldParser,
    VersionNineFieldParser,
    VersionOneFieldParser,
    VersionSevenFieldParser,
    VersionSixFieldParser,
    VersionTenFieldParser,
    VersionThreeFieldParser,
    VersionTwelveFieldParser,
    VersionTwoFieldParser,
)


class LicenseParser:
    def __init__(self, data: str) -> None:
        self._regex = Regex()
        self.data = self._clean_and_format_string(data)
        self.field_parser: FieldParser = FieldParser(self.data)

    def _clean_and_format_string(self, data: str) -> str:
        data = data.replace("\u001e", "").replace("\r", "")
        lines = data.split("\n")
        cleaned_lines = [line.strip() for line in lines if line.strip()]
        data = "\n".join(cleaned_lines)
        if not data.startswith("@"):
            data = "@\n" + data
        return data

    def parse(self) -> ParsedLicense:
        version = self.parse_version()
        self.field_parser = self._version_based_field_parsing(version)

        license_data: dict[str, object] = {
            "first_name": self.field_parser.parse_first_name(),
            "last_name": self.field_parser.parse_last_name(),
            "middle_name": self.field_parser.parse_middle_name(),
            "expiration_date": self.field_parser.parse_expiration_date(),
            "issue_date": self.field_parser.parse_issue_date(),
            "date_of_birth": self.field_parser.parse_date_of_birth(),
            "gender": self.field_parser.parse_gender(),
            "eye_color": self.field_parser.parse_eye_color(),
            "height": self.field_parser.parse_height(),
            "street_address": self.field_parser.parse_string("street_address"),
            "city": self.field_parser.parse_string("city"),
            "state": self.field_parser.parse_string("state"),
            "postal_code": self.field_parser.parse_string("postal_code"),
            "drivers_license_id": self.field_parser.parse_string("drivers_license_id"),
            "document_id": self.field_parser.parse_string("document_id"),
            "country": self.field_parser.parse_country(),
            "middle_name_truncation": self.field_parser.parse_truncation_status(
                "middle_name_truncation"
            ),
            "first_name_truncation": self.field_parser.parse_truncation_status(
                "first_name_truncation"
            ),
            "last_name_truncation": self.field_parser.parse_truncation_status(
                "last_name_truncation"
            ),
            "street_address_supplement": self.field_parser.parse_string(
                "street_address_supplement"
            ),
            "hair_color": self.field_parser.parse_hair_color(),
            "place_of_birth": self.field_parser.parse_string("place_of_birth"),
            "audit_information": self.field_parser.parse_string("audit_information"),
            "inventory_control_number": self.field_parser.parse_string("inventory_control_number"),
            "last_name_alias": self.field_parser.parse_string("last_name_alias"),
            "first_name_alias": self.field_parser.parse_string("first_name_alias"),
            "suffix_alias": self.field_parser.parse_string("suffix_alias"),
            "suffix": self.field_parser.parse_name_suffix(),
            "version": version,
            "pdf417": self.data,
            "expired": self.field_parser.parse_is_expired(),
            "weight": self.field_parser.parse_string("weight"),
            "cdl_indicator": self.field_parser.parse_string("cdl_indicator"),
            "non_domiciled_indicator": self.field_parser.parse_string("non_domiciled_indicator"),
            "enhanced_credential_indicator": self.field_parser.parse_string(
                "enhanced_credential_indicator"
            ),
            "permit_indicator": self.field_parser.parse_string("permit_indicator"),
        }

        return License.from_dict(license_data)

    def parse_version(self) -> str | None:
        return self._regex.first_match(r"\d{6}(\d{2})\w+", self.data)

    def _version_based_field_parsing(self, version: str | None) -> FieldParser:
        default_parser = FieldParser(self.data)

        if not version:
            return default_parser

        parsers: dict[str, type[FieldParser]] = {
            "01": VersionOneFieldParser,
            "02": VersionTwoFieldParser,
            "03": VersionThreeFieldParser,
            "04": VersionFourFieldParser,
            "05": VersionFiveFieldParser,
            "06": VersionSixFieldParser,
            "07": VersionSevenFieldParser,
            "08": VersionEightFieldParser,
            "09": VersionNineFieldParser,
            "10": VersionTenFieldParser,
            "11": VersionElevenFieldParser,
            "12": VersionTwelveFieldParser,
        }
        parser_cls = parsers.get(version)
        return parser_cls(self.data) if parser_cls else default_parser

    def is_expired(self) -> bool:
        expiration_date = self.field_parser.parse_expiration_date()
        return expiration_date is not None and datetime.now() > expiration_date



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime
from typing import Any, cast

from aamva_parser.enums import (
    EyeColor,
    Gender,
    HairColor,
    IssuingCountry,
    NameSuffix,
    Truncation,
)


@dataclass(slots=True)
class License:
    """Decoded AAMVA PDF417 record (one card scan)."""

    first_name: str | None = None
    last_name: str | None = None
    middle_name: str | None = None
    expiration_date: datetime | None = None
    issue_date: datetime | None = None
    date_of_birth: datetime | None = None
    gender: Gender = Gender.UNKNOWN
    eye_color: EyeColor = EyeColor.UNKNOWN
    height: float | None = None
    street_address: str | None = None
    city: str | None = None
    state: str | None = None
    postal_code: str | None = None
    drivers_license_id: str | None = None
    document_id: str | None = None
    country: IssuingCountry = IssuingCountry.UNKNOWN
    middle_name_truncation: Truncation = Truncation.NONE
    first_name_truncation: Truncation = Truncation.NONE
    last_name_truncation: Truncation = Truncation.NONE
    street_address_supplement: str | None = None
    hair_color: HairColor = HairColor.UNKNOWN
    place_of_birth: str | None = None
    audit_information: str | None = None
    inventory_control_number: str | None = None
    last_name_alias: str | None = None
    first_name_alias: str | None = None
    suffix_alias: str | None = None
    suffix: NameSuffix = NameSuffix.UNKNOWN
    cdl_indicator: str | None = None
    non_domiciled_indicator: str | None = None
    enhanced_credential_indicator: str | None = None
    permit_indicator: str | None = None
    version: str | None = None
    pdf417: str | None = None
    expired: bool = False
    weight: str | None = None

    @classmethod
    def from_dict(cls, data: Mapping[str, object]) -> License:
        known = {f.name for f in cls.__dataclass_fields__.values()}
        extra = set(data) - known
        if extra:
            raise TypeError(f"Unknown fields for {cls.__name__}: {sorted(extra)!s}")
        filtered = {k: v for k, v in data.items() if k in known}
        return cls(**cast(dict[str, Any], filtered))

    def is_expired(self) -> bool:
        return self.expiration_date is not None and datetime.now() > self.expiration_date

    def has_been_issued(self) -> bool:
        return self.issue_date is not None and datetime.now() > self.issue_date

    def is_acceptable(self) -> bool:
        return (
            not self.is_expired()
            and self.has_been_issued()
            and self.expiration_date is not None
            and self.last_name is not None
            and self.first_name is not None
            and self.middle_name is not None
            and self.issue_date is not None
            and self.date_of_birth is not None
            and self.height is not None
            and self.street_address is not None
            and self.city is not None
            and self.state is not None
            and self.postal_code is not None
            and self.document_id is not None
        )



from aamva_parser.regex_util import Regex


def test_first_match_returns_captured_group() -> None:
    regex = Regex()
    result = regex.first_match(r"hello (\w+)", "hello world")
    assert result == "world"


def test_first_match_returns_none_when_no_match() -> None:
    regex = Regex()
    result = regex.first_match(r"hello (\d+)", "hello world")
    assert result is None
